unit engine;

interface
uses objects, drivers, views, menus, app, dialogs, dos, crt,
     msgbox, stddlg, strutils, globals;
{$I decls.h}

var pwmcounter : byte;   { accessed with sim552 for determing pwmx outputs }

procedure ExecuteInstruction;
procedure ResetRegisters;
function  DecodeInstruction( CmdLine : String; InstrPtr : Integer; newpc : word ) : String;

implementation

const
  bitmask : array[0..7] of byte = ( $01, $02, $04, $08, $10, $20, $40, $80 );
  {TCON BitAddresses}
  IT0BitAddress = $88;
  IE0BitAddress = $89;
  IT1BitAddress = $8A;
  IE1BitAddress = $8B;
  TR0BitAddress = $8C;
  TF0BitAddress = $8D;
  TR1BitAddress = $8E;
  TF1BitAddress = $8F;
  {S0CON BitAddresses}
  RIBitAddress  = $98;
  TIBitAddress  = $99;
  RB8BitAddress = $9A;
  TB8BitAddress = $9B;
  RENBitAddress = $9C;
  SM2BitAddress = $9D;
  SM1BitAddress = $9E;
  SM0BitAddress = $9F;
  {IEN0 BitAddresses}
  EX0BitAddress = $A8;
  ET0BitAddress = $A9;
  EX1BitAddress = $AA;
  ET1BitAddress = $AB;
  ES0BitAddress = $AC;
  ES1BitAddress = $AD;
  EADBitAddress = $AE;
  EABitAddress  = $AF;
  {IP0 BitAddresses}
  PX0BitAddress = $B8;
  PT0BitAddress = $B9;
  PX1BitAddress = $BA;
  PT1BitAddress = $BB;
  PS0BitAddress = $BC;
  PS1BitAddress = $BD;
  PADBitAddress = $BE;
  {TM2IR BitAddresses}
  CTI0BitAddress = $C8;
  CTI1BitAddress = $C9;
  CTI2BitAddress = $CA;
  CTI3BitAddress = $CB;
  CMI0BitAddress = $CC;
  CMI1BitAddress = $CD;
  CMI2BitAddress = $CE;
  T2OVBitAddress = $CF;
  {PSW BitAddresses}
  PBitAddress   = $D0;
  F1BitAddress  = $D1;
  OVBitAddress  = $D2;
  RS0BitAddress = $D3;
  RS1BitAddress = $D4;
  F0BitAddress  = $D5;
  ACBitAddress  = $D6;
  CYBitAddress  = $D7;
  {S1CON BitAddresses}
  CR0BitAddress = $D8;
  CR1BitAddress = $D9;
  AABitAddress  = $DA;
  SIBitAddress  = $DB;
  STOBitAddress = $DC;
  STABitAddress = $DD;
  ENS1BitAddress= $DE;
  {IEN1 BitAddresses}
  ECT0BitAddress= $E8;
  ECT1BitAddress= $E9;
  ECT2BitAddress= $EA;
  ECT3BitAddress= $EB;
  ECM0BitAddress= $EC;
  ECM1BitAddress= $ED;
  ECM2BitAddress= $EE;
  ET2BitAddress = $EF;
  {IP1 BitAddresses}
  PCT0BitAddress= $F8;
  PCT1BitAddress= $F9;
  PCT2BitAddress= $FA;
  PCT3BitAddress= $FB;
  PCM0BitAddress= $FC;
  PCM1BitAddress= $FD;
  PCM2BitAddress= $FE;
  PT2BitAddress = $FF;

var
  oldpwmp,      { keeps value of PWMP }
  pwmprescaler, { prescaler value of PWMP }
  fosc          { keeps track of cycles for pwmp calculations }
   : byte;

{ returns state of bit from a bit address in IDATA }
function GetIBit( BitAddress : Byte ) : boolean;
var ea, ival : byte;
begin
   ea := $20 + (BitAddress DIV 8);     {00-07 = $20, 08-0f = $21 etc     }
   BitAddress := BitAddress AND $0f;   { mask low nibble                 }
   if BitAddress > $07 then BitAddress := BitAddress - $08;
   { now bitaddress has a value between 0 and 7, and ea is the iram byte }
   ival := iram[ea] AND bitmask[bitaddress];
   if ival > 0 then GetIBit := true else GetIBit := false;
end;

{ clears bit from a bit address in IDATA }
procedure ClearIBit( BitAddress : Byte );
var ea, ival : byte;
begin
   {00-07 = $20, 08-0f = $21 etc}
   ea := $20 + (BitAddress DIV 8);
   BitAddress := BitAddress AND $0f;
   if BitAddress > $07 then BitAddress := BitAddress - $08;
   { now bitaddress has a value between 0 and 7, and ea is the iram byte }
   iram[ea] := iram[ea] AND ($FF - bitmask[bitaddress]);
end;

{ sets bit from a bit address in IDATA }
procedure SetIBit( BitAddress : Byte );
var ea, ival : byte;
begin
   {00-07 = $20, 08-0f = $21 etc}
   ea := $20 + (BitAddress DIV 8);
   BitAddress := BitAddress AND $0f;
   if BitAddress > $07 then BitAddress := BitAddress - $08;
   { now bitaddress has a value between 0 and 7, and ea is the iram byte }
   iram[ea] := iram[ea] OR bitmask[bitaddress];
end;

{ returns state of bit from a bit address in sfr }
function GetSFRBit( BitAddress : Byte ) : boolean;
var ea, ival : byte;
begin
   ea := (BitAddress AND $f0) + (((BitAddress AND $0f) DIV 8) * 8);
   BitAddress := BitAddress AND $0f;
   if BitAddress > $07 then BitAddress := BitAddress - $08;
   { now bitaddress has a value between 0 and 7, and ea is the iram byte }
   ival := sfrram[ea] AND bitmask[bitaddress];
   if ival > 0 then GetSFRBit := true else GetSFRBit := false;
end;

{ clears bit from a bit address in sfr }
procedure ClearSFRBit( BitAddress : Byte );
var ea, ival : byte;
begin
   ea := (BitAddress AND $f0) + (((BitAddress AND $0f) DIV 8) * 8);
   BitAddress := BitAddress AND $0f;
   if BitAddress > $07 then BitAddress := BitAddress - $08;
   { now bitaddress has a value between 0 and 7, and ea is the iram byte }
   sfrram[ea] := sfrram[ea] AND ($FF - bitmask[bitaddress]);
end;

{ sets bit from a bit address in sfr }
procedure SetSFRBit( BitAddress : Byte );
var ea, ival : byte;
begin
   ea := (BitAddress AND $f0) + (((BitAddress AND $0f) DIV 8) * 8);
   BitAddress := BitAddress AND $0f;
   if BitAddress > $07 then BitAddress := BitAddress - $08;
   { now bitaddress has a value between 0 and 7, and ea is the iram byte }
   sfrram[ea] := sfrram[ea] OR bitmask[bitaddress];
end;

{ this is used to determine parity bit of ACC, as MOD 2 does not work }
function Parity( OpByte : Byte ) : boolean;
var bits, loop : byte;
begin
    bits := 0;
    for loop := 0 to 7 do
       if( OpByte AND BitMask[loop]) = BitMask[loop]  then
           bits := bits + 1;
    if bits in [1,3,5,7] then
       parity := true
    else
       parity := false;
end;

{ this checks if there will be a carry out of bit 7, and sets C if so }
procedure UpdateCarry( opcode1, opcode2 : byte );
var temp : word;
begin
    temp := opcode1 + opcode2;
    if temp > $ff then
       SetSFRBit(CYBitAddress)
    else
       ClearSFRBit(CYBitAddress);
end;

{ this checks if there will be a carry out of bit 3, and sets AC if so }
procedure UpdateHalfCarry( opcode1, opcode2 : byte );
var temp : word;
begin
    temp := (opcode1 AND $0f) + (opcode2 AND $0f);
    if temp > $0f then
       SetSFRBit(ACBitAddress)
    else
       ClearSFRBit(ACBitAddress);
end;

{ this checks if there will be a overflow out of bit 6, and sets C if so }
procedure UpdateOverflow( opcode1, opcode2 : byte );
var temp : word;
    carry6, carry7 : boolean;
begin
    { if carry from 6 but not 7, OR if carry from 7 but not 6, PSW.2 = 1 }
    temp := opcode1 + opcode2;
    if temp > $ff then
       carry7 := true
    else
       carry7 := false;
    temp := (opcode1 AND $4f) + (opcode2 AND $4f);
    if temp > $7f then
       carry6 := true
    else
       carry6 := false;
    if (carry6 AND (NOT carry7)) OR ((NOT carry6) AND carry7) then
       SetSFRBit(OVBitAddress)
    else
       ClearSFRBit(OVBitAddress);
end;

function NewPCVal( reloffset : byte; oldpc : word ) : word;
var offset : word;
begin
    offset := word( reloffset );
    if offset < 128 then NewPCVal := oldpc + offset
    else if offset > 127 then begin
       offset := 256 - offset;
       NewPCVal := oldpc - offset;
    end;
end;

procedure Increment( OpByte : Byte );
var ea, tmp1, carry : byte;
begin
     case OpByte of
          $04 : begin { INC A }
                   sfrram[ACC] := sfrram[ACC] + 1;
                end;
          $05 : begin { INC dir }
                   ea := cram[PC+1];
                   if ea > $7f then
                      sfrram[ea] := sfrram[ea] +1
                   else
                      iram[ea] := iram[ea] + 1;
                end;
          $06 : begin { inc @R0 }
                   ea := iram[regbank * 8];
                   iram[ea] := iram[ea] + 1;
                end;
          $07 :  begin { inc @R1 }
                   ea := iram[(regbank * 8) + 1];
                   iram[ea] := iram[ea] + 1;
                end;
          $08..$0F : begin { Inc R0..R7 }
                   tmp1 := OpByte - $08;
                   ea := (regbank * 8) + tmp1;
                   iram[ea] := iram[ea] + 1;
                end;
          $A3 : begin { inc dptr }
                   sfrram[DPL] := sfrram[DPL] + 1;
                   if sfrram[DPL] = $00 then carry := 1 else carry := 0;
                   sfrram[DPH] := sfrram[DPH] + carry;
                 end;
     end;
     PC := PC + Instrs[OpByte].nbytes;
end;

procedure Decrement( OpByte : Byte );
var ea, tmp1, carry : byte;
begin
     case OpByte of
          $14 : begin { DEC A }
                   sfrram[ACC] := sfrram[ACC] - 1;
                end;
          $15 : begin { DEC dir }
                   ea := cram[PC+1];
                   if ea > $7f then
                      sfrram[ea] := sfrram[ea] - 1
                   else
                      iram[ea] := iram[ea] - 1;
                end;
          $16 : begin { dec @R0 }
                   ea := iram[regbank * 8];
                   iram[ea] := iram[ea] - 1;
                end;
          $17 :  begin { dec @R1 }
                   ea := iram[(regbank * 8) + 1];
                   iram[ea] := iram[ea] - 1;
                end;
          $18..$1F : begin { DEC R0..R7 }
                   tmp1 := OpByte - $18;
                   ea := (regbank * 8) + tmp1;
                   iram[ea] := iram[ea] - 1;
                end;
     end;
     PC := PC + Instrs[OpByte].nbytes;
end;

procedure Add( OpByte : Byte );
var ea, tmp1, op1, op2 : byte;
begin
     op1 := sfrram[ACC];
     case OpByte of
          $24 : begin { ADD A, #DATA }
                   op2 := cram[PC+1];
                   sfrram[ACC] := sfrram[ACC] + cram[PC+1];
                end;
          $25 : begin { ADD A,dir }
                   ea  := cram[PC+1];
                   op2 := iram[ea];
                   if ea > $7f then
                      sfrram[ACC] := sfrram[ACC] + sfrram[ea]
                   else
                      sfrram[ACC] := sfrram[ACC] + iram[ea]
                end;
          $26 : begin { ADD A, @R0 }
                   ea := iram[regbank * 8];
                   op2 := iram[ea];
                   sfrram[ACC] := sfrram[ACC] + iram[ea];
                end;
          $27 : begin { ADD A, @R1 }
                   ea := iram[(regbank * 8) + 1];
                   op2 := iram[ea];
                   sfrram[ACC] := sfrram[ACC] + iram[ea];
                end;
          $28..$2F : begin { ADD A, R0..R7 }
                   tmp1 := OpByte - $28;
                   ea := (regbank * 8) + tmp1;
                   op2 := iram[ea];
                   sfrram[ACC] := sfrram[ACC] + iram[ea];
                end;
     end;
     UpdateCarry( op1, op2 );
     UpdateHalfCarry( op1, op2 );
     UpdateOverflow( op1, op2 );
     PC := PC + Instrs[OpByte].nbytes;
end;

procedure AddWithCarry( OpByte : Byte );
var ea, tmp1, op1, op2, carry : byte;
begin
     op1   := sfrram[ACC];
     if GetSFRBit(CYBitAddress) then carry := $01 else carry := $00;
     case OpByte of
          $34 : begin { ADDC A, #DATA }
                   op2 := cram[PC+1] + carry;
                   sfrram[ACC] := sfrram[ACC] + cram[PC+1] + carry;
                end;
          $35 : begin { ADDC A,dir }
                   ea  := cram[PC+1];
                   op2 := iram[ea] + carry;
                   if ea > $7f then
                      sfrram[ACC] := sfrram[ACC] + sfrram[ea] + carry
                   else
                      sfrram[ACC] := sfrram[ACC] + iram[ea] + carry;
                end;
          $36 : begin { ADDC A, @R0 }
                   ea := iram[regbank * 8];
                   op2 := iram[ea] + carry;
                   sfrram[ACC] := sfrram[ACC] + iram[ea] + carry;
                end;
          $37 : begin { ADDC A, @R1 }
                   ea := iram[(regbank * 8) + 1];
                   op2 := iram[ea] + carry;
                   sfrram[ACC] := sfrram[ACC] + iram[ea] + carry;
                end;
          $38..$3F : begin { ADDC A, R0..R7 }
                   tmp1 := OpByte - $38;
                   ea := (regbank * 8) + tmp1;

                   op2 := iram[ea] + carry;
                   sfrram[ACC] := sfrram[ACC] + iram[ea] + carry;
                end;
     end;
     UpdateCarry( op1, op2 );
     UpdateHalfCarry( op1, op2 );
     UpdateOverflow( op1, op2 );
     PC := PC + Instrs[OpByte].nbytes;
end;

procedure SubTractWithBorrow( OpByte : Byte );
var ea, tmp1, op1, op2, carry, borrow7, borrow6, borrow3 : byte;
begin
     op1   := sfrram[ACC];
     borrow7 := $00;
     borrow6 := $00;
     borrow3 := $00;
     if GetSFRBit(CYBitAddress) then carry := $01 else carry := $00;
     case OpByte of
          $94 : begin { SUBB A, #DATA }
                   op2 := cram[PC+1] + carry;
                   sfrram[ACC] := sfrram[ACC] - carry - cram[PC+1];
                end;
          $95 : begin { SUBB A,dir }
                   ea  := cram[PC+1];
                   op2 := iram[ea] + carry;
                   if ea > $7f then
                      sfrram[ACC] := sfrram[ACC] - carry - sfrram[ea]
                   else
                      sfrram[ACC] := sfrram[ACC] - carry - iram[ea];
                end;
          $96 : begin { SUBB A, @R0 }
                   ea := iram[regbank * 8];
                   op2 := iram[ea] + carry;
                   sfrram[ACC] := sfrram[ACC] - carry - iram[ea];
                end;
          $97 : begin { SUBB A, @R1 }
                   ea := iram[(regbank * 8) + 1];
                   op2 := iram[ea] + carry;
                   sfrram[ACC] := sfrram[ACC] - carry - iram[ea];
                end;
          $98..$9F : begin { SUBB A, R0..R7 }
                   tmp1 := OpByte - $98;
                   ea := (regbank * 8) + tmp1;
                   op2 := iram[ea] + carry;
                   sfrram[ACC] := sfrram[ACC] - carry - iram[ea];
                end;
     end;
     { op1 = ACC, op2 = data}
     if op1 < op2 then borrow7 := $01;
     if (op1 AND $0f) < (op2 AND $0f) then borrow3 := $01;
     if (op1 AND $7f) < (op2 AND $7f) then borrow6 := $01;
     if borrow7 = $01 then SetSFRBit( CYBitAddress )
     else ClearSFRBit( CYBitAddress );
     if borrow3 = $01 then SetSFRBit( ACBitAddress )
     else ClearSFRBit( ACBitAddress );
     if (borrow6 = $01) AND (borrow7 = $00) then
        SetSFRBit( OvBitAddress );
     if (borrow6 = $00) AND (borrow7 = $01) then
        SetSFRBit( OvBitAddress );
     PC := PC + Instrs[OpByte].nbytes;
end;

procedure AbsoluteCall( OpByte : Byte );
var  ea1, ea2, temppc : word;
begin
     temppc := pc + 2;
     sfrram[SP] := sfrram[SP]+1;  {increment sp}
     ea1 := sfrram[SP];
     iram[ea1] := temppc AND $00ff;  { store PC7-0 }
     sfrram[SP] := sfrram[SP]+1;  {increment sp}
     ea1 := sfrram[SP];
     iram[ea1] := ((temppc AND $FF00) SHR 8) AND $00FF;  { store PC15-8 }
     ea1 := PC + 2;
     { high order 5bits of PC + bits 7-5 of opcode + second byte }
     ea2 := ((OpByte AND $E0) SHL 3) AND $fff7;
     PC := (ea1 AND $F800) + ea2 + cram[PC+1];
end;

procedure AbsoluteJump( OpByte : Byte );
var  ea1, ea2 : word;
begin
    ea1 := PC + 2;
    { high order 5bits of PC + bits 7-5 of opcode + second byte }
    ea2 := ((OpByte AND $E0) SHL 3) AND $fff7;
    PC := (ea1 AND $F800) + ea2 + cram[PC+1];
end;

procedure LongJumpCall( OpByte : Byte );
var  ea1, ea2, temppc : word;
begin
     case OpByte of
        $02 : begin { LJMP addr16 }
                 ea1 := cram[PC+1];
                 ea1 := (ea1 SHL 8) AND $FF00;
                 ea2 := cram[PC+2];
                 PC := ea1 + ea2;
              end;
        $12 : begin { LCALL addr16 }
                 temppc := pc + 3;
                 sfrram[SP] := sfrram[SP]+1;  {increment sp}
                 ea1 := sfrram[SP];
                 iram[ea1] := temppc AND $00ff;  { store PC7-0 }
                 sfrram[SP] := sfrram[SP]+1;  {increment sp}
                 ea1 := sfrram[SP];
                 iram[ea1] := ((temppc AND $FF00) SHR 8) AND $00FF;  { store PC15-8 }
                 ea1 := (cram[PC+1] SHL 8) AND $FF00;
                 ea2 := cram[PC+2];
                 PC := ea1 + ea2;
              end;
     end;
end;

procedure PushPop( OpByte : Byte );
var  ea1, ea2 : byte;
begin
     case OpByte of
        $C0 : begin { PUSH dir }
                 sfrram[SP] := sfrram[SP]+1;  {increment sp}
                 ea1 := sfrram[SP];
                 ea2 := cram[PC+1];
                 if ea2 > $7f then
                    iram[ea1] := sfrram[ea2]
                 else
                    iram[ea1] := iram[ea2];
              end;
        $D0 : begin { POP dir }
                 ea1 := sfrram[SP];
                 ea2 := cram[PC+1];
                 if ea2 > $7f then
                    sfrram[ea2] := iram[ea1]
                 else
                    iram[ea2] := iram[ea1];
                 sfrram[SP] := sfrram[SP]-1;  {decrement sp}
              end;
     end;
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure Return( OpByte : Byte );
var  ea : byte;
     temppc : word;
begin
     case OpByte of
        $22 : begin { RET }
                 ea := sfrram[SP];
                 temppc := iram[ea];
                 temppc := (temppc SHL 8) AND $FF00;
                 sfrram[SP] := sfrram[SP]-1;  {decrement sp}
                 ea := sfrram[SP];
                 PC := temppc + iram[ea];
              end;
        $32 : begin { RETI }
                 ea := sfrram[SP];
                 temppc := iram[ea];
                 temppc := (temppc SHL 8) AND $FF00;
                 sfrram[SP] := sfrram[SP]-1;  {decrement sp}
                 ea := sfrram[SP];
                 PC := temppc + iram[ea];
              end;
     end;
end;

procedure AndLogical( OpByte : Byte );
var ea, tmp1 : byte;
    bitstate : boolean;
begin
    case OpByte of
         $52 : begin { ANL dir, A }
                  ea := cram[PC+1];
                  if ea > $7f then
                     sfrram[ea] := sfrram[ea] AND sfrram[ACC]
                  else
                     iram[ea] := iram[ea] AND sfrram[ACC];
               end;
         $53 : begin { ANL dir, #data }
                  ea := cram[PC+1];
                  if ea > $7f then
                     sfrram[ea] := sfrram[ea] AND cram[PC+2]
                  else
                     iram[ea] := iram[ea] AND cram[PC+2];
               end;
         $54 : begin { ANL A, #data }
                  sfrram[ACC] := sfrram[ACC] AND cram[PC+1];
               end;
         $55 : begin { ANL A, dir }
                  ea := cram[PC+1];
                  if ea > $7f then
                     sfrram[ACC] := sfrram[ACC] AND sfrram[ea]
                  else
                     sfrram[ACC] := sfrram[ACC] AND iram[ea];
               end;
         $56 : begin { ANL A, @R0 }
                  ea := iram[regbank * 8];
                  sfrram[ACC] := sfrram[ACC] AND iram[ea];
               end;
         $57 :  begin { ANL A, @R1 }
                  ea := iram[(regbank * 8) + 1];
                  sfrram[ACC] := sfrram[ACC] AND iram[ea];
               end;
         $58..$5F : begin { ANL A, R0..R7 }
                  tmp1 := OpByte - $58;
                  ea := (regbank * 8) + tmp1;
                  sfrram[ACC] := sfrram[ACC] AND iram[ea];
               end;
         $82: begin { ANL C, bit }
                  ea := cram[PC+1];
                  if ea > $7f then
                     bitstate := GetSFRBit( ea )
                  else
                     bitstate := GetIBit( ea );
                  if NOT bitstate then
                     ClearSFRBit(CYBitAddress);
               end;
         $B0 : begin { ANL C, /bit }
                  ea := cram[PC+1];
                  if ea > $7f then
                     bitstate := GetSFRBit( ea )
                  else
                     bitstate := GetIBit( ea );
                  if bitstate then
                     ClearSFRBit(CYBitAddress);
               end;
    end;
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure OrLogical( OpByte : Byte );
var ea, tmp1 : byte;
    bitstate : boolean;
begin
    case OpByte of
         $42 : begin { ORL dir, A }
                  ea := cram[PC+1];
                  if ea > $7f then
                     sfrram[ea] := sfrram[ea] OR sfrram[ACC]
                  else
                     iram[ea] := iram[ea] OR sfrram[ACC];
               end;
         $43 : begin { ORL dir, #data }
                  ea := cram[PC+1];
                  if ea > $7f then
                     sfrram[ea] := sfrram[ea] OR cram[PC+2]
                  else
                     iram[ea] := iram[ea] OR cram[PC+2];
               end;
         $44 : begin { ORL A, #data }
                  sfrram[ACC] := sfrram[ACC] OR cram[PC+1];
               end;
         $45 : begin { ORL A, dir }
                  ea := cram[PC+1];
                  if ea > $7f then
                     sfrram[ACC] := sfrram[ACC] OR sfrram[ea]
                  else
                     sfrram[ACC] := sfrram[ACC] OR iram[ea];
               end;
         $46 : begin { ORL A, @R0 }
                  ea := iram[regbank * 8];
                  sfrram[ACC] := sfrram[ACC] OR iram[ea];
               end;
         $47 :  begin { ORL A, @R1 }
                  ea := iram[(regbank * 8) + 1];
                  sfrram[ACC] := sfrram[ACC] OR iram[ea];
               end;
         $48..$4F : begin { ORL A, R0..R7 }
                  tmp1 := OpByte - $48;
                  ea := (regbank * 8) + tmp1;
                  sfrram[ACC] := sfrram[ACC] OR iram[ea];
               end;
         $72: begin { ORL C, bit }
                  ea := cram[PC+1];
                  if ea > $7f then
                     bitstate := GetSFRBit( ea )
                  else
                     bitstate := GetIBit( ea );
                  if bitstate then
                     SetSFRBit(CYBitAddress);
               end;
         $A0 : begin { ORL C, /bit }
                  ea := cram[PC+1];
                  if ea > $7f then
                     bitstate := GetSFRBit( ea )
                  else
                     bitstate := GetIBit( ea );
                  if NOT bitstate then
                     SetSFRBit(CYBitAddress);
               end;
    end;
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure XorLogical( OpByte : Byte );
var ea, tmp1 : byte;
    bitstate : boolean;
begin
    case OpByte of
         $62 : begin { XRL dir, A }
                  ea := cram[PC+1];
                  if ea > $7f then
                     sfrram[ea] := sfrram[ea] XOR sfrram[ACC]
                  else
                     iram[ea] := iram[ea] XOR sfrram[ACC];
               end;
         $63 : begin { XRL dir, #data }
                  ea := cram[PC+1];
                  if ea > $7f then
                     sfrram[ea] := sfrram[ea] XOR cram[PC+2]
                  else
                     iram[ea] := iram[ea] XOR cram[PC+2];
               end;
         $64 : begin { XRL A, #data }
                  sfrram[ACC] := sfrram[ACC] XOR cram[PC+1];
               end;
         $65 : begin { XRL A, dir }
                  ea := cram[PC+1];
                  if ea > $7f then
                     sfrram[ACC] := sfrram[ACC] XOR sfrram[ea]
                  else
                     sfrram[ACC] := sfrram[ACC] XOR iram[ea];
               end;
         $66 : begin { XRL A, @R0 }
                  ea := iram[regbank * 8];
                  sfrram[ACC] := sfrram[ACC] XOR iram[ea];
               end;
         $67 :  begin { XRL A, @R1 }
                  ea := iram[(regbank * 8) + 1];
                  sfrram[ACC] := sfrram[ACC] XOR iram[ea];
               end;
         $68..$6F : begin { XRL A, R0..R7 }
                  tmp1 := OpByte - $68;
                  ea := (regbank * 8) + tmp1;
                  sfrram[ACC] := sfrram[ACC] XOR iram[ea];
               end;
    end;
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure RotateA( OpByte : Byte );
var bit0, bit7, tmp : byte;
begin
    case OpByte of
         $03 : begin { RR A }
                  bit0 := ((sfrram[ACC] AND $01) SHL 7) AND $80;
                  tmp := ((sfrram[ACC] SHR 1) AND $7f) + bit0;
                  sfrram[ACC] := tmp;
               end;
         $13 : begin { RRC A }
                  if GetSFRBit( CYBitAddress ) then bit7 := $80
                  else bit7 := $00;
                  bit0 := sfrram[ACC] AND $01;
                  if bit0 = $01 then SetSFRBit( CYBitAddress )
                  else ClearSFRBit( CYBitAddress );
                  tmp := ((sfrram[ACC] SHR 1) AND $7f) + bit7;
                  sfrram[ACC] := tmp;
               end;
         $23 : begin { RL A }
                  bit7 := ((sfrram[ACC] AND $80) SHR 7) AND $01;
                  tmp := ((sfrram[ACC] SHL 1) AND $fe) + bit7;
                  sfrram[ACC] := tmp;
               end;
         $33 :  begin { RLC A }
                  if GetSFRBit( CYBitAddress ) then bit0 := $01
                  else bit0 := $00;
                  bit7 := sfrram[ACC] AND $80;
                  if bit7 = $80 then SetSFRBit( CYBitAddress )
                  else ClearSFRBit( CYBitAddress );
                  tmp := ((sfrram[ACC] SHL 1) AND $fe) + bit0;
                  sfrram[ACC] := tmp;
               end;
    end;
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure Clear( OpByte : Byte );
var ea : byte;
begin
    case OpByte of
         $E4 : begin { CLR A }
                  sfrram[ACC] := $00;
               end;
         $C3 : begin { CLR C }
                  ClearSFRBit( CyBitAddress );
               end;
         $C2 : begin { CLR bit }
                  ea := cram[PC+1];
                  if ea > $7f then
                     ClearSFRBit( ea )
                  else
                     ClearIBit( ea );
               end;
    end;
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure CompareAndJump( OpByte : Byte );
var ea1, tmp1, op1, op2 : byte;
    temppc, ea : word;
begin
    temppc := PC + 3;
    op1 := cram[PC+1];
    op2 := cram[PC+2];  { relative address }
    case OpByte of
         $B4 : begin { CJNE A, #DATA, rel }
                  if op1 < sfrram[ACC] then begin
                     ea := NewPCVal( op2, temppc );
                     ClearSFRBit(CYBitAddress);
                  end
                  else if op1 > sfrram[ACC] then begin
                     ea := NewPCVal( op2, temppc );
                     SetSFRBit(CYBitAddress);
                  end
                  else begin
                     ClearSFRBit(CYBitAddress);
                     ea := PC + 3;
                  end;
               end;
         $B5 : begin { CJNE A, dir, rel }
                  if op1 > $7f then
                     tmp1 := sfrram[op1]
                  else
                     tmp1 := iram[op1];
                  if tmp1 < sfrram[ACC] then begin
                     ea := NewPCVal( op2, temppc );
                     ClearSFRBit(CYBitAddress);
                  end
                  else if tmp1 > sfrram[ACC] then begin
                     ea := NewPCVal( op2, temppc );
                     SetSFRBit(CYBitAddress);
                  end
                  else begin
                     ClearSFRBit(CYBitAddress);
                     ea := PC + 3;
                  end;
               end;
         $B6 : begin { CJNE @R0, #data, rel }
                  ea1 := iram[regbank * 8];
                  tmp1 := iram[ea1];
                  if op1 < tmp1 then begin
                     ea := NewPCVal( op2, temppc );
                     ClearSFRBit(CYBitAddress);
                  end
                  else if op1 > tmp1 then begin
                     ea := NewPCVal( op2, temppc );
                     SetSFRBit(CYBitAddress);
                  end
                  else begin
                     ClearSFRBit(CYBitAddress);
                     ea := PC + 3;
                  end;
               end;
         $B7 : begin { CJNE @R1, #data, rel }
                  ea1 := iram[(regbank * 8) + 1];
                  tmp1 := iram[ea1];
                  if op1 < tmp1 then begin
                     ea := NewPCVal( op2, temppc );
                     ClearSFRBit(CYBitAddress);
                  end
                  else if op1 > tmp1 then begin
                     ea := NewPCVal( op2, temppc );
                     SetSFRBit(CYBitAddress);
                  end
                  else begin
                     ClearSFRBit(CYBitAddress);
                     ea := PC + 3;
                  end;
               end;
         $B8..$BF : begin { CJNE R0..R7, #data, rel }
                  tmp1 := OpByte - $B8;
                  ea1 := (regbank * 8) + tmp1;
                  tmp1 := iram[ea1];
                  if op1 < tmp1 then begin
                     ea := NewPCVal( op2, temppc );
                     ClearSFRBit(CYBitAddress);
                  end
                  else if op1 > tmp1 then begin
                     ea := NewPCVal( op2, temppc );
                     SetSFRBit(CYBitAddress);
                  end
                  else begin
                     ClearSFRBit(CYBitAddress);
                     ea := PC + 3;
                  end;
               end;
    end;
    PC := ea;
end;

procedure DecJumpNotZero( OpByte : Byte );
var ea1, tmp1, op1, op2 : byte;
    temppc, ea : word;
begin
    temppc := PC + 2;
    op1 := cram[PC+1];
    op2 := cram[PC+2];  { relative address }
    case OpByte of
         $D5 : begin { DNJZ dir, rel }
                  temppc := pc + 3;
                  if op1 > $7f then begin
                     ea1 := sfrram[op1] - 1;
                     sfrram[op1] := ea1;
                  end
                  else begin
                     ea1 := iram[op1] - 1;
                     iram[op1] := ea1;
                  end;
                  if ea1 <> 0 then
                     ea := NewPCVal( op2, temppc )
                  else
                     ea := temppc;
               end;
         $D8..$DF : begin { DJNZ Rn, rel }
                  tmp1 := OpByte - $D8;
                  ea1 := (regbank * 8) + tmp1;
                  tmp1 := iram[ea1] - 1;
                  iram[ea1] := tmp1;
                  if tmp1 <> 0 then
                     ea := NewPCVal( op1, temppc )
                  else
                     ea := temppc;
               end;
    end;
    PC := ea;
end;

procedure Complement( OpByte : Byte );
var ea, tmp1 : byte;
    bitstate : boolean;
begin
    case OpByte of
         $F4 : begin { CPL A }
                  sfrram[ACC] := $FF - sfrram[ACC];
               end;
         $B3 : begin { CPL C }
                  if GetSFRBit(CYBitAddress) then
                     ClearSFRBit(CYBitAddress)
                  else
                     SetSFRBit(CYBitAddress);
               end;
         $B2 : begin { CPL bit }
                  ea := cram[PC+1];
                  if ea > $7f then begin
                     if GetSFRBit( ea ) then
                        ClearSFRBit( ea )
                     else
                        SetSFRBit( ea );
                  end
                  else begin
                     if GetIBit( ea ) then
                        ClearIBit( ea )
                     else
                        SetIBit( ea );
                  end;
               end;
    end;
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure DivAB( OpByte : Byte );
var quotient, remainder : byte;
begin
    if sfrram[B] = $00 then
    begin
        setSFRbit( OVBitAddress );
    end
    else begin
        quotient := sfrram[ACC] DIV sfrram[B];
        remainder:= sfrram[ACC] MOD sfrram[B];
        sfrram[ACC] := quotient;
        sfrram[B]   := remainder;
        ClearSFRBit( OVBitAddress );
    end;
    ClearSFRBit( CYBitAddress );
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure MulAB( OpByte : Byte );
var product : word;
begin
    product := word(sfrram[ACC]) * word(sfrram[B]);
    sfrram[ACC] := product AND $00ff;
    sfrram[B] := (product SHR 8) AND $00ff;
    if product > $00ff then
       SetSFRBit( OVBitAddress );
    ClearSFRBit( CYBitAddress );
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure SetB( OpByte : Byte );
var ea : byte;
begin
    case OpByte of
         $D2 : begin { SETB bit }
                  ea := cram[PC+1];
                  if ea > $7f then
                     SetSFRBit( ea )
                  else
                     SetIBit( ea );
               end;
         $D3 : begin { SETB C }
                   SetSFRBit(CYBitAddress);
               end;
    end;
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure MoveBit( OpByte : Byte );
var ea : byte;
    bitstate : boolean;
begin
    case OpByte of
         $92 : begin { MOV Bit, C }
                  ea := cram[PC+1];
                  bitstate := GetSFRBit( CYBitAddress );
                  if bitstate then begin
                     if ea > $7f then
                        SetSFRBit( ea )
                     else
                        SetIBit( ea )
                  end
                  else begin
                     if ea > $7f then
                        ClearSFRBit( ea )
                     else
                        ClearIBit( ea );
                  end;
               end;
         $A2 : begin { MOV C, bit }
                  ea := cram[PC+1];
                  if ea > $7f then
                     bitstate := GetSFRBit( ea )
                  else
                     bitstate := GetIBit( ea );
                  if bitstate then
                      SetSFRBit( CYBitAddress )
                   else
                      ClearSFRBit( CYBitAddress );
               end;
    end;
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure MoveImmediateData( OpByte : Byte );
var ea, tmp1 : byte;
begin
    case OpByte of
         $74 : begin { MOV A, #data }
                  sfrram[ACC] := cram[PC+1];
               end;
         $75 : begin { MOV dir, #data }
                  ea := cram[PC+1];
                  if ea > $7f then
                     sfrram[ea] := cram[PC+2]
                  else
                     iram[ea] := cram[PC+2];
               end;
         $76 : begin { MOV @R0, #data }
                  ea := iram[(regbank * 8)];
                  iram[ea] := cram[PC+1];
               end;
         $77 :  begin { MOV @R1, #data }
                  ea := iram[(regbank * 8) + 1];
                  iram[ea] := cram[PC+1];
               end;
         $78..$7F : begin { MOV R0..R7, #data }
                  tmp1 := OpByte - $78;
                  ea := (regbank * 8) + tmp1;
                  iram[ea] := cram[PC+1];
               end;
    end;
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure MoveDirect( OpByte : Byte );
var ea, tmp1 : byte;
begin
    case OpByte of
         $85 : begin { MOV dir, dir }
                  ea := cram[PC+1];
                  tmp1 := cram[PC+2];
                  if ea > $7f then begin
                    if tmp1 > $7f then
                       sfrram[ea] := sfrram[tmp1]
                    else
                       sfrram[ea] := iram[tmp1];
                  end
                  else begin
                    if tmp1 > $7f then
                       iram[ea] := sfrram[tmp1]
                    else
                       iram[ea] := iram[tmp1];
                  end;
               end;
         $86 : begin { MOV dir, @R0 }
                  ea := iram[(regbank * 8)];
                  tmp1 := cram[PC+1];
                  if tmp1 > $7f then
                     sfrram[tmp1] := iram[ea]
                  else
                     iram[tmp1] := iram[ea];
               end;
         $87 :  begin { MOV dir, @R1 }
                  ea := iram[(regbank * 8) + 1];
                  tmp1 := cram[PC+1];
                  if tmp1 > $7f then
                     sfrram[tmp1] := iram[ea]
                  else
                     iram[tmp1] := iram[ea];
               end;
         $88..$8F : begin { MOV dir, R0..R7 }
                  tmp1 := OpByte - $88;
                  ea := (regbank * 8) + tmp1;
                  tmp1 := cram[PC+1];
                  if tmp1 > $7f then
                     sfrram[tmp1] := iram[ea]
                  else
                     iram[tmp1] := iram[ea];
               end;
    end;
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure MoveRegisterDirect( OpByte : Byte );
var ea, tmp1 : byte;
begin
    case OpByte of
         $A6 : begin { MOV @R0, dir }
                  ea := iram[(regbank * 8)];
                  tmp1 := cram[PC+1];
                  if tmp1 > $7f then
                     iram[ea] := sfrram[tmp1]
                  else
                     iram[ea] := iram[tmp1];
               end;
         $A7 :  begin { MOV @R1, dir }
                  ea := iram[(regbank * 8) + 1];
                  tmp1 := cram[PC+1];
                  if tmp1 > $7f then
                     iram[ea] := sfrram[tmp1]
                  else
                     iram[ea] := iram[tmp1];
               end;
         $A8..$AF : begin { MOV R0..R7, dir }
                  tmp1 := OpByte - $A8;
                  ea := (regbank * 8) + tmp1;
                  tmp1 := cram[PC+1];
                  if tmp1 > $7f then
                     iram[ea] := sfrram[tmp1]
                  else
                     iram[ea] := iram[tmp1];
               end;
    end;
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure MoveACCRegister( OpByte : Byte );
var ea, tmp1 : byte;
begin
    case OpByte of
         $E5 : begin { MOV A, dir }
                  ea := cram[PC+1];
                  if ea > $7f then
                     sfrram[ACC] := sfrram[ea]
                  else
                     sfrram[ACC] := iram[ea];
               end;
         $E6 : begin { MOV A, @R0 }
                  ea := iram[(regbank * 8)];
                  sfrram[ACC] := iram[ea];
               end;
         $E7 :  begin { MOV A, @R1 }
                  ea := iram[(regbank * 8) + 1];
                  sfrram[ACC] := iram[ea];
               end;
         $E8..$EF : begin { MOV A, R0..R7 }
                  tmp1 := OpByte - $E8;
                  ea := (regbank * 8) + tmp1;
                  sfrram[ACC] := iram[ea];
               end;
    end;
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure MoveRegisterACC( OpByte : Byte );
var ea, tmp1 : byte;
begin
    case OpByte of
         $F5 : begin { MOV dir, A }
                  ea := cram[PC+1];
                  if ea > $7f then
                     sfrram[ea] := sfrram[ACC]
                  else
                     iram[ea] := sfrram[ACC];
               end;
         $F6 : begin { MOV @R0, A }
                  ea := iram[(regbank * 8)];
                  iram[ea] := sfrram[ACC];
               end;
         $F7 :  begin { MOV @R1, A }
                  ea := iram[(regbank * 8) + 1];
                  iram[ea] := sfrram[ACC];
               end;
         $F8..$FF : begin { MOV R0..R7, A }
                  tmp1 := OpByte - $F8;
                  ea := (regbank * 8) + tmp1;
                  iram[ea] := sfrram[ACC];
               end;
    end;
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure MoveX( OpByte : Byte );
var ea, tmp1 : byte;
    ea1 : word;
begin
    case OpByte of
         $E0 : begin { MOVX A, @DPTR }
                  ea1 := ((sfrram[DPH] SHL 8) AND $ff00)+sfrram[DPL];
                  sfrram[ACC] := cram[ea1];
               end;
         $E2 : begin { MOVX A, @R0 }
                  ea := iram[(regbank * 8)];
                  sfrram[ACC] := xxram[ea];
               end;
         $E3 :  begin { MOVX A, @R1 }
                  ea := iram[(regbank * 8) + 1];
                  sfrram[ACC] := xxram[ea];
               end;
         $F0 : begin { MOVX @DPTR, A }
                  ea1 := ((sfrram[DPH] SHL 8) AND $ff00)+sfrram[DPL];
                  xxram[ea1] := sfrram[ACC];
               end;
         $F2 : begin { MOVX @R0, A }
                  ea := iram[(regbank * 8)];
                  xxram[ea] := sfrram[ACC];
               end;
         $F3 :  begin { MOVX @R1, A }
                  ea := iram[(regbank * 8) + 1];
                  xxram[ea] := sfrram[ACC];
               end;
    end;
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure MoveCodeByte( OpByte : Byte );
var ea, tmp1 : byte;
    ea1 : word;
begin
    case OpByte of
         $83 : begin { MOVC A, A+PC }
                  ea1 := sfrram[ACC] + PC + 1;
                  sfrram[ACC] := cram[ea1];
               end;
         $93 : begin { MOVC A, A+@DPTR }
                  ea1 := ((sfrram[DPH] SHL 8) AND $ff00)+sfrram[DPL] + sfrram[ACC];
                  sfrram[ACC] := cram[ea1];
               end;
    end;
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure MoveDPTRImmediate( OpByte : Byte );
begin
    sfrram[DPH] := cram[PC+1];
    sfrram[DPL] := cram[PC+2];
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure JumpADPTR( OpByte : Byte );
var ea1 : word;
begin
    ea1 := ((sfrram[DPH] SHL 8) AND $ff00)+sfrram[DPL];
    ea1 := ea1 + sfrram[ACC];
    PC := ea1;
end;

procedure SwapA( OpByte : Byte );
var lnib, hnib : byte;
begin
    lnib := ((sfrram[ACC] AND $0f) SHL 4) AND $f0;
    hnib := ((sfrram[ACC] AND $f0) SHR 4) AND $0f;
    sfrram[ACC] := lnib + hnib;
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure JumpBit( OpByte : Byte );
var ea : byte;
    temppc : word;
    bitstate : boolean;
begin
    case OpByte of
         $10 : begin { JBC bit, rel }
                  temppc := PC;
                  ea := cram[PC+1];
                  if ea > $7f then
                     Bitstate := GetSFRBit( ea )
                  else
                     bitstate := GetIBit( ea );
                  if bitstate then
                     PC := NewPCVal( cram[PC+2], PC+3 )
                  else
                     PC := PC+3;
                  if ea > $7f then
                     ClearSFRBit( ea )
                  else
                     ClearIBit( ea );
               end;
         $20 : begin { JB bit, rel }
                  ea := cram[PC+1];
                  if ea > $7f then
                     Bitstate := GetSFRBit( ea )
                  else
                     bitstate := GetIBit( ea );
                  if bitstate then
                     PC := NewPCVal( cram[PC+2], PC+3 )
                  else
                     PC := PC+3;
               end;
         $30 : begin { JNB bit, rel }
                  ea := cram[PC+1];
                  if ea > $7f then
                     Bitstate := GetSFRBit( ea )
                  else
                     bitstate := GetIBit( ea );
                  if bitstate then
                     PC := PC+3
                  else
                     PC := NewPCVal( cram[PC+2], PC+3 );
               end;
         $40 : begin { JC, rel }
                  if GetSFRBit( CYBitAddress ) then
                     PC := NewPCVal( cram[PC+1], PC+2 )
                  else
                     PC := PC+2;
               end;
         $50 : begin { JNC, rel }
                  if GetSFRBit( CYBitAddress ) then
                     PC := PC + 2
                  else
                     PC := NewPCVal( cram[PC+1], PC+2 );
               end;
         $60 : begin { JZ, rel }
                  if sfrram[ACC] = $00 then
                     PC := NewPCVal( cram[PC+1], PC+2 )
                  else
                     PC := PC + 2;
               end;
         $70 : begin { JNZ, rel }
                  if sfrram[ACC] <> $00 then
                     PC := NewPCVal( cram[PC+1], PC+2 )
                  else
                     PC := PC + 2;
               end;
         $80 : begin { SJMP rel }
                  PC := NewPCVal( cram[PC+1], PC+2 )
               end;
    end;
end;

procedure XChange( OpByte : Byte );
var ea, tmp1, ac : byte;
begin
    ac := sfrram[ACC];
    case OpByte of
         $C5 : begin { XCH A, dir }
                  ea := cram[PC+1];
                  if ea > $7f then begin
                     sfrram[ACC] := sfrram[ea];
                     sfrram[ea] := ac;
                  end
                  else begin
                     sfrram[ACC] := iram[ea];
                     iram[ea] := ac;
                  end;
               end;
         $C6 : begin { XCH A, @R0 }
                  ea := iram[(regbank * 8)];
                  sfrram[ACC] := iram[ea];
                  iram[ea] := ac;
               end;
         $C7 :  begin { XCH A, @R1 }
                  ea := iram[(regbank * 8) + 1];
                  sfrram[ACC] := iram[ea];
                  iram[ea] := ac;
               end;
         $C8..$CF : begin { XCH A, R0..R7 }
                  tmp1 := OpByte - $C8;
                  ea := (regbank * 8) + tmp1;
                  sfrram[ACC] := iram[ea];
                  iram[ea] := ac;
               end;
         $D6 : begin { XCHD A, @R0 }
                  ea := iram[(regbank * 8)];
                  sfrram[ACC] := (ac AND $f0) + (iram[ea] AND $0f);
                  iram[ea] := (iram[ea] AND $f0) + (ac AND $0f);
               end;
         $D7 :  begin { XCH A, @R1 }
                  ea := iram[(regbank * 8) + 1];
                  sfrram[ACC] := (ac AND $f0) + (iram[ea] AND $0f);
                  iram[ea] := (iram[ea] AND $f0) + (ac AND $0f);
               end;
    end;
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure DAA( OpByte : Byte );
var carry, halfcarry : boolean;
    low, high : byte;
begin
    low := sfrram[ACC] AND $0f;
    high := ((sfrram[ACC] AND $f0) SHR 4) AND $0f;
    carry := GetSFRBit( CYBitAddress );
    halfcarry := GetSFRBit( ACBitAddress );
    if (low > 9) OR halfcarry then begin
       low := low + 6;
       carry := true;
    end;
    if (high > 9) OR carry then begin
       high := high + 6;
       carry := true;
    end;
    sfrram[ACC] := ((high SHL 4) AND $f0) + low;
    if carry then SetSFRBit( CYBitAddress );
    PC := PC + Instrs[OpByte].nbytes;
end;

procedure ExecuteInstruction;
var OpByte : Byte;
begin
    {use PC value, decode instruction, adjust cycles for clock count}
    OpByte := cram[PC];
    { temporary insert for regbank for testing purposes }
    regbank := ((sfrram[PSW] AND $18) SHR 3) AND $03;
    case OpByte of
         $00 : begin { NOP }
                   PC := PC + Instrs[OpByte].nbytes;
                end;
         $01, $21, $41, $61, $81, $A1, $C1, $E1 : begin { AJMP addr11 }
                   AbsoluteJump( OpByte );
                end;
         $02, $12 : begin { LJMP addr16,  LCALL addr16 }
                   LongJumpCall( OpByte );
                end;
         $03, $13, $23, $33 : begin { RR A, RRC A, RL A, RLC A }
                   RotateA( OpByte );
                end;
         $04..$0F, $A3 : begin { INC }
                   Increment( OpByte );
                end;
         $10, $20, $30, $40, $50, $60, $70, $80 : begin { JumpCond bit, rel }
                   JumpBit( OpByte );
                end;
         $11, $31, $51, $71, $91, $B1, $D1, $F1 : begin { ACALL }
                   AbsoluteCall( OpByte );
                end;
         $14..$1F : begin { DEC }
                   Decrement( OpByte );
                end;
         $22, $32 : begin { RET,  RETI }
                   Return( OpByte );
                end;
         $24..$2F : begin { ADD }
                   Add( OpByte );
                end;
         $34..$3F : begin { ADDC }
                   AddWithCarry( OpByte );
                end;
         $42..$4F, $72, $A0 : begin { ORL A,   ORL C, }
                   OrLogical( OpByte );
                end;
         $52..$5F, $82, $B0 : begin { ANL A,   ANL C, }
                   AndLogical( OpByte );
                end;
         $62..$6F : begin { XRL A,   XRL C, }
                   XorLogical( OpByte );
               end;
         $73 : begin { JMP @A+DPTR }
                   JumpADPTR( OpByte );
               end;
         $74..$7F : begin { MOV .., #Data }
                   MoveImmediateData( OpByte );
               end;
         $83, $93 : Begin { MOVC A, @A+ }
                   MoveCodeByte( OpByte );
               end;
         $84 : Begin { DIV AB }
                   DivAB( OpByte );
               end;
         $85..$8F : begin { MOV direct, ... }
                   MoveDirect( OpByte );
               end;
         $90 : begin { MOV DPTR, #data16 }
                   MoveDPTRImmediate( OpByte );
               end;
         $92, $A2 : begin { Mov bit,c  Mov c, bit }
                  MoveBit( OpByte );
                end;
         $94..$9F : begin { ADDC }
                   SubtractWithBorrow( OpByte );
                end;
         $A4 : Begin { MUL AB }
                   MulAB( OpByte );
               end;
         $A6..$AF : begin { MOV Rx, direct }
                   MoveRegisterDirect( OpByte );
               end;
         $B2, $B3, $F4 : begin { CPL }
                   Complement( OpByte );
                end;
         $B4..$BF : begin { CJNE }
                   CompareAndJump( OpByte );
                end;
         $C0, $D0 : begin { PUSH dir, POP dir }
                   PushPop( OpByte );
                end;
         $C2, $C3, $E4 : begin { CLR A,   CLR C, }
                   Clear( OpByte );
                end;
         $C4 : Begin { SWAP A }
                   SwapA( OpByte );
               end;
         $C5..$CF, $D6, $D7 : begin { XCH }
                   XChange( OpByte );
                end;
         $D2, $D3 : begin { SETB bit, SETB C }
                   SetB( OpByte );
                end;
         $D4 : begin { DAA }
                   DAA( OpByte );
               end;
         $D5, $D8..$DF : begin { DJNZ }
                   DecJumpNotZero( OpByte );
               end;
         $E0, $E2, $E3, $F0, $F2, $F3 : begin { MOVX }
                   MoveX( OpByte );
               end;
         $E5..$EF : begin { MOV Rx, direct }
                   MoveACCRegister( OpByte );
               end;
         $F5..$FF : begin { MOV Rx, direct }
                   MoveRegisterACC( OpByte );
               end;
         else
                   PC := PC + Instrs[OpByte].nbytes;
    end;
    { update clock cycles }
    fosc := fosc + Instrs[OpByte].Cycles;
    Cycles := Cycles + Instrs[OpByte].Cycles;
    { update pwmcounter, clock rate / (2x(1+sfrram[pwmp]x255) }
    if oldpwmp <> sfrram[PWMP] then begin { PWMP has been changed         }
       oldpwmp := sfrram[PWMP];   { load new value to check for next time }
       pwmprescaler := sfrram[PWMP]; { reset prescaler value to new value }
    end;
    while ( fosc >= 2 ) do begin
       { decrement prescaler every 1/2 clock rate}
       pwmprescaler := pwmprescaler - 1;
       fosc := fosc - 2;
    end;
    if pwmprescaler = $00 then begin { if prescaler = 0 then inc counter  }
       pwmprescaler := oldpwmp;      { and reload prescaler               }
       pwmcounter := pwmcounter + 1;
       if pwmcounter = $ff then pwmcounter := 0; { modulo 254             }
    end;
    { adjust parity bit to reflect contents of ACC }
    if parity(sfrram[ACC]) then
       SetSFRBit(PBitAddress)
    else
       ClearSFRBit(PBitAddress);
    { range check PC }
    if PC >= xramlimit then Running80C552 := false;
end;

procedure ResetRegisters;
begin
     sfrram[ACC]   := $00;
     sfrram[ADCON] := sfrram[ADCON] AND $C0;
     sfrram[B]     := $00;
     sfrram[CML0]  := $00;
     sfrram[CML1]  := $00;
     sfrram[CML2]  := $00;
     sfrram[CMH0]  := $00;
     sfrram[CMH1]  := $00;
     sfrram[CMH2]  := $00;
     sfrram[CTCON] := $00;
     sfrram[DPL]   := $00;
     sfrram[DPH]   := $00;
     sfrram[IEN0]  := $00;
     sfrram[IEN1]  := $00;
     sfrram[IP0]   := sfrram[IP0] AND $80;
     sfrram[IP1]   := $00;
     PC := $0000;
     sfrram[PCON]  := sfrram[PCON] AND $60;
     sfrram[PSW]   := $00;
     sfrram[PWM0]  := $00;
     sfrram[PWM1]  := $00;
     sfrram[PWMP]  := $00;
     sfrram[P0]    := $FF;
     sfrram[P1]    := $FF;
     sfrram[P2]    := $FF;
     sfrram[P3]    := $FF;
     sfrram[P4]    := $FF;
     sfrram[RTE]   := $00;
     sfrram[S0CON] := $00;
     sfrram[S1ADR] := $00;
     sfrram[S1CON] := sfrram[S1CON] AND $80;
     sfrram[S1DAT] := $00;
     sfrram[S1STA] := $F8;
     sfrram[SP]    := $07;
     sfrram[STE]   := $C0;
     sfrram[TCON]  := $00;
     sfrram[TH0]   := $00;
     sfrram[TH1]   := $00;
     sfrram[TMH2]  := $00;
     sfrram[TL0]   := $00;
     sfrram[TL1]   := $00;
     sfrram[TML2]  := $00;
     sfrram[TMOD]  := $00;
     sfrram[TM2CON]:= $00;
     sfrram[TM2IR] := $00;
     sfrram[T3]    := $00;
end;

function DecodeInstruction( CmdLine : String; InstrPtr : Integer; newpc : Word ) : String;
var  Instruction : String;
     OPByte      : Integer;
     ea1, ea2, destpc : word;
     destadr : string;
begin
   OpByte := StrtoInteger(CmdLine[InstrPtr] + CmdLine[InstrPtr+1]);
   Instruction := Instrs[OpByte].Name;
   {add and replace XX's where appropriate }
   case Instrs[OpByte].adrmode of
      bit7, dir7, predir : begin
                             Instruction[7] := CmdLine[Instrptr+2];
                             Instruction[8] := CmdLine[InstrPtr+3];
                          end;
      addr16    : begin
                             Instruction[7] := CmdLine[Instrptr+2];
                             Instruction[8] := CmdLine[InstrPtr+3];
                             Instruction[9] := CmdLine[Instrptr+4];
                             Instruction[10]:= CmdLine[InstrPtr+5];
                          end;
      addr11    : begin
                             ea1 := newpc + 2;
                             ea2 := ((OpByte AND $E0) SHL 3) AND $fff7;
                             destpc := (ea1 AND $f800) + ea2 + cram[newpc+1];
                             destadr := WordToString(destpc);
                             Instruction[7] := destadr[1];
                             Instruction[8] := destadr[2];
                             Instruction[9] := destadr[3];
                             Instruction[10]:= destadr[4];
                          end;
      dir10, data8, bit10 : begin
                             Instruction[10] := CmdLine[Instrptr+2];
                             Instruction[11] := CmdLine[InstrPtr+3];
                          end;
      rel10 : begin { djnz rn, rel }
                             destpc := NewPCVal( cram[newpc+1], newpc+2 );
                             destadr := WordToString(destpc);
                             Instruction[10] := destadr[1];
                             Instruction[11] := destadr[2];
                             Instruction[12] := destadr[3];
                             Instruction[13] := destadr[4];
                          end;
      dirdata, dirdir : begin
                             Instruction[7]  := CmdLine[Instrptr+2];
                             Instruction[8]  := CmdLine[InstrPtr+3];
                             Instruction[11] := CmdLine[Instrptr+4];
                             Instruction[12] := CmdLine[InstrPtr+5];
                          end;
      rel : begin { Jx rel,  Sjmp rel }
                             destpc := NewPCVal( cram[newpc+1], newpc+2 );
                             destadr := WordToString(destpc);
                             Instruction[7] := destadr[1];
                             Instruction[8] := destadr[2];
                             Instruction[9] := destadr[3];
                             Instruction[10] := destadr[4];
                          end;
      bitrel : begin  {JBx XX,XX}
                             Instruction[7]  := CmdLine[Instrptr+2];
                             Instruction[8]  := CmdLine[InstrPtr+3];
                             destpc := NewPCVal( cram[newpc+2], newpc+3 );
                             destadr := WordToString(destpc);
                             Instruction[11] := destadr[1];
                             Instruction[12] := destadr[2];
                             Instruction[13] := destadr[3];
                             Instruction[14] := destadr[4];
                          end;
      datarel, dirrel :   begin { cjne xx, rel}
                             Instruction[11] := CmdLine[Instrptr+2];
                             Instruction[12] := CmdLine[InstrPtr+3];
                             destpc := NewPCVal( cram[newpc+2], newpc+3 );
                             destadr := WordToString(destpc);
                             Instruction[14] := destadr[1];
                             Instruction[15] := destadr[2];
                             Instruction[16] := destadr[3];
                             Instruction[17] := destadr[4];
                          end;
      dirrel7 :           begin {djnz xx, rel}
                             Instruction[7]  := CmdLine[Instrptr+2];
                             Instruction[8]  := CmdLine[InstrPtr+3];
                             destpc := NewPCVal( cram[newpc+2], newpc+3 );
                             destadr := WordToString(destpc);
                             Instruction[10] := destadr[1];
                             Instruction[11] := destadr[2];
                             Instruction[12] := destadr[3];
                             Instruction[13] := destadr[4];
                          end;
      data16 :            begin
                             Instruction[11] := CmdLine[Instrptr+2];
                             Instruction[12] := CmdLine[InstrPtr+3];
                             Instruction[13] := CmdLine[Instrptr+4];
                             Instruction[14] := CmdLine[InstrPtr+5];
                          end;
      crel :              begin {MOVC  A,@A+XXXX}
                             destadr := WordToString(newpc+1+sfrram[ACC]);
                             Instruction[15] := destadr[1];
                             Instruction[16] := destadr[2];
                             Instruction[17] := destadr[3];
                             Instruction[18] := destadr[4];
                          end;
   end;
   DecodeInstruction := Instruction;
end;

begin
   fosc := 0;
   pwmprescaler := sfrram[PWMP];
   oldpwmp := sfrram[PWMP];
   pwmcounter := 0;
end.
