
unit strutils; { strutils.pas   a unit of string conversion routines }


INTERFACE

const
     hexdigits : array[0..15] of char = ( '0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F' );

     binvalues : array[0..15] of string[4] = (
          ( '0000' ), ( '0001' ), ( '0010' ), ( '0011' ),
          ( '0100' ), ( '0101' ), ( '0110' ), ( '0111' ),
          ( '1000' ), ( '1001' ), ( '1010' ), ( '1011' ),
          ( '1100' ), ( '1101' ), ( '1110' ), ( '1111' ));

function integertostring( item : integer ) : string;
function wordtostring( item : word ) : string;
function strtobin( by : string ) : string;
function strtobyte( by : string ) : byte;
function strtointeger( by : string ) : integer;
function strtoword( by : string ) : word;
function strtoascii( by : string ) : string;
procedure StrUpper( var st : string );
function ValidHexStr( HexStr : String ) : Boolean;


implementation
var st : string;
    len, num : integer;

function integertostring( item : integer ) : string;
begin
   { convert integer to two byte hex string with leading zero }
   num := item;
   st := '';
   while num > 15 do begin
      st  := st + hexdigits[num DIV 16];
      num := num MOD 16;
   end;
   st := st + hexdigits[num];
   while length(st) < 2 do st := '0' + st;
   integertostring := st
end;

function wordtostring( item : word ) : string;
begin
   { convert word to four byte hex string with leading zeros }
   st := integertostring( item SHR 8 );
   st := st + integertostring( item AND $00ff );
   wordtostring := st
end;

function strtobin( by : string ) : string;
begin
   { convert by to binary string of 8 of 01 }
   { by input is 2 characters long }
   st := '';
   if by[1] in ['A'..'F','a'..'f'] then
      st := binvalues[(ord(by[1]) - $37)]
   else
      st := binvalues[ord(by[1]) - $30];
   if by[2] in ['A'..'F','a'..'f'] then
      st := st + binvalues[(ord(by[2]) - $37)]
   else
      st := st + binvalues[ord(by[2]) - $30];
   strtobin := st;
end;

function strtobyte( by : string ) : byte;
var num : byte;
begin
   {two hex bytes to convert}
   while length(by) < 2 do
      by := '0' + by;
   if by[1] in ['A'..'F','a'..'f'] then
      num := (ord(by[1]) - $37) * 16
   else
      num := ((ord(by[1]) - $30) * 16);
   if by[2] in ['A'..'F','a'..'f'] then
      num := num + (ord(by[2]) - $37)
   else
      num := num + (ord(by[2]) - $30);
   strtobyte := num;
end;

function strtointeger( by : string ) : integer;
begin
   {two hex bytes to convert}
   while length(by) < 2 do
      by := '0' + by;
   if by[1] in ['A'..'F','a'..'f'] then
      num := (ord(by[1]) - $37) * 16
   else
      num := ((ord(by[1]) - $30) * 16);
   if by[2] in ['A'..'F','a'..'f'] then
      num := num + (ord(by[2]) - $37)
   else
      num := num + (ord(by[2]) - $30);
   strtointeger := num;
end;

function strtoword( by : string ) : word;
var num : word;
begin
   {four hex bytes to convert}
   while length(by) < 4 do
      by := '0' + by;
   num := 0;
   if by[1] in ['A'..'F','a'..'f'] then
      num := num + (ord(by[1]) - $37) * 16 * 16 * 16
   else
      num := num + ((ord(by[1]) - $30) * 16 * 16 * 16);
   if by[2] in ['A'..'F','a'..'f'] then
      num := num + ((ord(by[2]) - $37) * 16 * 16)
   else
      num := num + ((ord(by[2]) - $30) * 16 * 16);
   if by[3] in ['A'..'F','a'..'f'] then
      num := num + (ord(by[3]) - $37) * 16
   else
      num := num + ((ord(by[3]) - $30) * 16);
   if by[4] in ['A'..'F','a'..'f'] then
      num := num + (ord(by[4]) - $37)
   else
      num := num + (ord(by[4]) - $30);
   strtoword := num;
end;

function strtoascii( by : string ) : string;
begin
   st := '';
   len := 1;
   while len < length(by) do begin
      num := strtointeger( by[len]+by[len+1] );
      if ord(num) in [32..127] then
         st := st + chr(ord(num))
      else
         st := st + '.';
      len := len + 2;
   end;
   strtoascii := st;
end;

procedure StrUpper( var st : string );
var  len, loop : integer;
     ch : byte;
begin
   len := length(st);
   for loop := 1 to len do
      if st[loop] in ['a'..'z'] then
         st[loop] := chr(ord(st[loop]) - ord('a') + ord('A'));
end;


function ValidHexStr( HexStr : String ) : Boolean;
var digit : integer;
    result: boolean;
begin
    result := true;
    for digit := 1 to length(HexStr) do
       if NOT (HexStr[digit] in ['0'..'9','A'..'F','a'..'f']) then
          result := false;
    ValidHexStr := result;
end;

begin
end.
